from django.shortcuts import render, redirect
from django_discord_connector.models import DiscordClient, DiscordToken, DiscordUser
from django.http import HttpResponseRedirect, HttpResponseForbidden, HttpResponse
from django.contrib import messages
from django.apps import apps
from requests_oauthlib import OAuth2Session
import logging
import base64
import requests


def sso_callback(request):
    try:
        discord_client = DiscordClient.get_instance()
    except:
        messages.warning(
            request, "The site administrator has not added the Discord Client to the admin panel.")
        return redirect('dashboard')
    data = {
        "client_id": discord_client.client_id,
        "client_secret": discord_client.client_secret,
        "grant_type": "authorization_code",
        "code": request.GET['code'],
        "redirect_uri": discord_client.callback_url,
    }
    headers = {
        "Content-Type": "application/x-www-form-urlencoded"
    }

    r = requests.post('%s/oauth2/token' %
                      discord_client.api_endpoint, data, headers)
    r.raise_for_status()

    json = r.json()
    token = json['access_token']
    me = requests.get('https://discordapp.com/api/users/@me',
                      headers={'Authorization': "Bearer " + token}).json()
    join = requests.post(discord_client.invite_link, headers={
                         'Authorization': "Bearer " + token}).json()
    # Catch errors
    if DiscordToken.objects.filter(discord_user__external_id=me['id']).exists():
        discord_token = DiscordToken.objects.get(discord_user__external_id=me['id'])
        if discord_token.user != request.user:
            messages.error(request, "That Discord user is claimed by another user.")
            return redirect('/')
    if not me['email']:
        messages.add_message(
            request, messages.ERROR, 'Could not find an email on your Discord profile. Please make sure your not signed in as a Guest Discord user.')
        return redirect('/')
    if me['email'] != request.user.email:
        messages.add_message(
            request, messages.WARNING, 'You linked a Discord account with a mismatched email, please verify you linked the correct Discord account.'
        )


    # Delete old token if exists
    if DiscordToken.objects.filter(user=request.user).exists():
        discord_token = request.user.discord_token
        discord_token.delete()

    # Get or Create Discord User
    discord_user = DiscordUser.objects.get_or_create(external_id=me['id'])[0]
    discord_user.username = me['username'] + "#" + me['discriminator']
    if 'nick' in me:
        discord_user.nickname = me['nick'] + "#" + me['discriminator']

    discord_user.save()

    # Attach DiscordToken to user
    token = DiscordToken(
        access_token=json['access_token'],
        refresh_token=json['refresh_token'],
        discord_user=discord_user,
        user=request.user
    )
    token.save()

    return redirect('/')


def add_sso_token(request):
    scope = (['email', 'guilds.join', 'identify'])
    try:
        discord_client = DiscordClient.get_instance()
    except:
        messages.warning(
            request, "Discord Settings are not configured correctly. Contact your administrator.")
        return redirect('/')
    oauth = OAuth2Session(discord_client.client_id,
                          redirect_uri=discord_client.callback_url, scope=scope, token=None, state=None)
    url, state = oauth.authorization_url(discord_client.base_uri)
    return HttpResponseRedirect(url)


def remove_sso_token(request, pk):
    discord_token = DiscordToken.objects.get(pk=pk)
    if request.user == discord_token.user:
        discord_token.delete()
        messages.add_message(request, messages.SUCCESS,
                             "Discord token has been deleted.")
        return redirect('/')
    else:
        messages.add_message(request, messages.ERROR,
                             "That Discord token does not belong to you.")
        return redirect('/')



import json
import requests
from discord.ext import commands
import logging


logger = logging.getLogger(__name__)


class DiscordRequest():
    def __init__(self, api_endpoint=None, server_id=None, client_id=None, secret=None, bot_token=None, settings=None):
        if settings:
            self.api_endpoint = str(settings['api_endpoint'])
            self.server_id = int(settings['server_id'])
            self.client_id = int(settings['client_id'])
            self.secret = str(settings['client_secret'])
            self.bot_token = str(settings['bot_token'])
        else:
            self.api_endpoint = api_endpoint
            self.server_id = server_id
            self.client_id = client_id
            self.secret = secret
            self.bot_token = bot_token

    @staticmethod
    def get_instance():
        from django_discord_connector.models import DiscordClient
        return DiscordRequest(settings=DiscordClient.get_instance().serialize())

    def activate(self):
        prefix = "?"
        bot = commands.Bot(command_prefix=prefix)
        bot.run(self.bot_token)

    def get_discord_user(self, user_id):
        url = self.api_endpoint + "/guilds/" + \
            str(self.server_id) + "/members/" + str(user_id)
        response = requests.get(url, headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Bot ' + self.bot_token
        })
        return response

    def get_discord_users(self):
        url = self.api_endpoint + "/guilds/" + self.server_id + "/members/"
        response = requests.get(url, headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Bot ' + self.bot_token
        })
        return response

    def add_role_to_user(self, role_id, user_id):
        url = self.api_endpoint + "/guilds/" + str(self.server_id) + \
            "/members/" + str(user_id) + "/roles/" + str(role_id)
        response = requests.put(url, headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Bot ' + self.bot_token
        })
        return response

    def remove_role_from_user(self, role_id, user_id):
        url = self.api_endpoint + "/guilds/" + str(self.server_id) + \
            "/members/" + str(user_id) + "/roles/" + str(role_id)
        response = requests.delete(url, headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Bot ' + self.bot_token
        })
        return response

    def add_role_to_server(self, role_name):
        url = self.api_endpoint + "/guilds/" + str(self.server_id) + "/roles"
        data = json.dumps({'name': role_name})
        response = requests.post(url,
                                 data=data,
                                 headers={
                                     'Content-Type': 'application/x-www-form-urlencoded',
                                     'Authorization': 'Bot ' + self.bot_token
                                 }
                                 )
        return response

    def remove_role_from_server(self, role_id):
        url = self.api_endpoint + "/guilds/" + \
            str(self.server_id) + "/roles/" + str(role_id)
        response = requests.delete(url, headers={
            'Content-Type': 'application/x-www-form-urlencoded',
            'Authorization': 'Bot ' + self.bot_token
        })
        return response

    def send_channel_message(self, channel_id, message=None, embed=None):
        url = self.api_endpoint + "/channels/" + str(channel_id) + "/messages"
        data = json.dumps({
            'content': message,
            'embed': embed, 
        })
        response = requests.post(url,
                                 data=data,
                                 headers={
                                     'Content-Type': 'application/json',
                                     'Authorization': 'Bot ' + self.bot_token
                                 }
                                 )
        return response

    def get_guild_roles(self):
        url = self.api_endpoint + "/guilds/" + str(self.server_id) + "/roles"
        response = requests.get(
            url,
            headers={
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': 'Bot ' + self.bot_token
            }
        )
        return response

    def get_guild_channels(self):
        url = self.api_endpoint + "/guilds/" + \
            str(self.server_id) + "/channels"
        response = requests.get(
            url,
            headers={
                'Content-Type': 'application/x-www-form-urlencoded',
                'Authorization': 'Bot ' + self.bot_token
            }
        )
        return response

    def update_discord_user_nickname(self, discord_user_id, discord_user_nickname):
        url = "%s/guilds/%s/members/%s" % (self.api_endpoint, str(self.server_id), discord_user_id)
        logger.debug("Calling Discord API: %s" % url)
        data = {
            "nick": discord_user_nickname,
        }
        response = requests.patch(
            url,
            headers={
                'Content-Type': 'application/json',
                'Authorization': 'Bot ' + self.bot_token
            },
            data=json.dumps(data)
        )
        return response

    def create_dm(self, recipient_id):
        url = "https://discord.com/api/users/@me/channels"
        logger.debug("Calling Discord API: %s" % url)
        data = {
            "recipient_id": recipient_id,
        }
        response = requests.patch(
            url,
            headers={
                'Content-Type': 'application/json',
                'Authorization': 'Bot ' + self.bot_token
            },
            data=json.dumps(data)
        )
        return response



# Generated by Django 2.2.4 on 2019-08-23 17:22

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('django_discord_connector', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='discordtoken',
            name='external_id',
        ),
        migrations.RemoveField(
            model_name='discordtoken',
            name='groups',
        ),
        migrations.RemoveField(
            model_name='discordtoken',
            name='nickname',
        ),
        migrations.RemoveField(
            model_name='discordtoken',
            name='username',
        ),
        migrations.CreateModel(
            name='DiscordUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(max_length=255)),
                ('nickname', models.CharField(max_length=255)),
                ('external_id', models.BigIntegerField()),
                ('groups', models.ManyToManyField(to='django_discord_connector.DiscordGroup')),
            ],
        ),
        migrations.AddField(
            model_name='discordtoken',
            name='discord_user',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='discord_token', to='django_discord_connector.DiscordUser'),
        ),
    ]



from django.contrib import admin
from django_discord_connector.models import DiscordClient, DiscordToken, DiscordGroup, DiscordChannel, DiscordUser
from django_discord_connector.tasks import sync_discord_groups
from django.apps import apps

def update_discord_groups(modeladmin, request, queryset):
    sync_discord_groups()
update_discord_groups.short_description="Update Discord groups from Discord Server"

# Register your models here.
admin.site.register(DiscordGroup)
admin.site.register(DiscordChannel)
admin.site.register(DiscordUser)

if apps.is_installed('django_singleton_admin'):
    # Highly Recommended: https://github.com/porowns/django-singleton-admin
    from django_singleton_admin.admin import DjangoSingletonModelAdmin
    @admin.register(DiscordClient)
    class DiscordClientAdmin(DjangoSingletonModelAdmin):
        fieldsets = (
            ('General Settings', {
                'fields': ('callback_url', 'server_id', 'client_id', 'client_secret', 'bot_token', 'invite_link' )
            }),
            ('Advanced Settings', {
                'classes': ('collapse', 'open'),
                'fields': ('api_endpoint', 'base_uri', 'token_uri', 'token_revoke_uri')
            }),
        )


from django.apps import AppConfig
from django.db.models.signals import m2m_changed, post_delete
from django.apps import apps

class DjangoDiscordConnectorConfig(AppConfig):
    name = 'django_discord_connector'
    package_name = __import__(name).__package_name__
    version = __import__(name).__version__
    verbose_name = 'discord'
    url_slug = 'discord'

    def ready(self):
        from .signals import user_group_change_sync_discord_groups, remove_discord_user_on_discord_token_removal
        from django.contrib.auth.models import User
        from django_discord_connector.models import DiscordToken
        m2m_changed.connect(
            user_group_change_sync_discord_groups, sender=User.groups.through)

        post_delete.connect(
            remove_discord_user_on_discord_token_removal,
            sender=DiscordToken
        )
        if apps.is_installed('packagebinder'):
            from .bindings import create_bindings
            create_bindings()


# Generated by Django 2.2.4 on 2019-09-04 16:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_discord_connector', '0004_auto_20190826_1348'),
    ]

    operations = [
        migrations.AlterField(
            model_name='discorduser',
            name='groups',
            field=models.ManyToManyField(blank=True, to='django_discord_connector.DiscordGroup'),
        ),
        migrations.AlterField(
            model_name='discorduser',
            name='nickname',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



from django.apps import apps 
from django.urls import reverse
from django.conf import settings
from .models import DiscordClient
from .forms import DiscordClientForm
from packagebinder.bind import PackageBinding, SettingsBinding, TaskBinding, SidebarBinding
import logging 

logger = logging.getLogger(__name__)

app_config = apps.get_app_config('django_discord_connector')

package_binding = PackageBinding(
    package_name=app_config.name, 
    version=app_config.version, 
    url_slug='discord', 
)

settings_binding = SettingsBinding(
    package_name=app_config.name, 
    settings_class=DiscordClient,
    settings_form=DiscordClientForm,
)

task_binding = TaskBinding(
    package_name=app_config.name, 
    required_tasks = [
        {
            "name": "Discord: Sync User Groups",
            "task_name": "django_discord_connector.tasks.verify_all_discord_users_groups",
            "interval": 5,
            "interval_period": "minutes",
        },
        {
            "name": "Discord: Update User Information",
            "task_name": "django_discord_connector.tasks.sync_all_discord_users_accounts",
            "interval": 1,
            "interval_period": "days",
        },
    ],
    optional_tasks = [
        {
            "name": "Discord: Hard Sync Users",
            "task_name": "django_discord_connector.tasks.remote_sync_all_discord_users_groups",
            "interval": 7,
            "interval_period": "days",
        },
    ]
)

def create_bindings():
    try:
        package_binding.save()
        settings_binding.save()
        task_binding.save()
    except Exception as e:
        if settings.DEBUG:
            raise(e)
        else:
            logger.error(f"Failed package binding step for {app_config.name}: {e}")



# Generated by Django 2.2.13 on 2020-11-20 21:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_discord_connector', '0009_auto_20200929_2007'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='discordclient',
            options={'verbose_name': 'Discord Settings', 'verbose_name_plural': 'Discord Settings'},
        ),
    ]



from django.dispatch import receiver
from django.db.models.signals import m2m_changed, post_delete
from django.contrib.auth.models import User
from .models import DiscordUser, DiscordGroup, DiscordToken
from .tasks import verify_discord_user_groups, remove_discord_user

import logging
logger = logging.getLogger(__name__)


@receiver(m2m_changed, sender=User.groups.through)
def user_group_change_sync_discord_groups(sender, **kwargs):
    django_user = kwargs.get('instance')
    action = str(kwargs.get('action'))

    # Check if Django user exists
    try:
        discord_user = DiscordUser.objects.get(discord_token__user=django_user)
    except DiscordUser.DoesNotExist:
        logger.info(
            "DiscordUser not found for %s, skipping group sync" % django_user)
        return

    if "post" in kwargs.get('action'):
        verify_discord_user_groups.apply_async(args=[discord_user.external_id], countdown=30)

@receiver(post_delete, sender=DiscordToken)
def remove_discord_user_on_discord_token_removal(sender, **kwargs):
    discord_token = kwargs.get('instance')
    try:
        remove_discord_user.apply_async(args=[discord_token.discord_user.external_id])
    except DiscordUser.DoesNotExist:
        return # dangling token 


from unittest.mock import patch
from django.test import TestCase
from django.contrib.auth.models import User, Group 
from django_discord_connector.models import DiscordUser, DiscordGroup, DiscordToken
import django_discord_connector
import logging 
logger = logging.getLogger('django_discord_connector')

def mock_discord_task(args, countdown=None):
    logger.info(f"args={args},countdown={countdown}")

class TestDiscordSignalSuite(TestCase):
    def setUp(self):
        self.group = Group.objects.create(
            name="Group"
        )

        self.user = User.objects.create(
            username="test"
        )

        self.discord_user = DiscordUser.objects.create(
            username="test",
            nickname="test",
            external_id=1,
        )

        self.discord_token = DiscordToken.objects.create(
            access_token="null",
            refresh_token="null",
            discord_user=self.discord_user,
            user=self.user
        )

        self.discord_group = DiscordGroup.objects.create(
            name="DiscordGroup",
            external_id=1,
            group=self.group
        )

    def tearDown(self):
        try:
            self.user.delete()
            self.group.delete()
            self.discord_token.delete()
            self.discord_user.delete()
            self.discord_group.delete()
        except Exception as e:
            pass

    @patch.object(django_discord_connector.signals.verify_discord_user_groups, 'apply_async', mock_discord_task)
    def test_user_group_change(self):
        with self.assertLogs('django_discord_connector', level='INFO') as cm:
            self.user.groups.add(self.group)
            message = cm.output[0]
            self.assertTrue("args=[1]" in message)
            self.assertTrue("countdown=30" in message)

    
    @patch.object(django_discord_connector.signals.remove_discord_user, 'apply_async', mock_discord_task)
    def test_user_delete(self):
        with self.assertLogs('django_discord_connector', level='INFO') as cm:
            self.user.delete()
            message = cm.output[0]
            self.assertTrue("args=[1]" in message)
        




from unittest.mock import patch
from django.test import TestCase, TransactionTestCase 
from django.contrib.auth.models import User, Group 
from django_discord_connector.models import DiscordUser, DiscordGroup, DiscordToken, DiscordClient
from django_discord_connector.tasks import *

class MockDiscordResponse():
    def __init__(self, status_code, response={}):
        self.status_code = status_code
        self.response = response

    def json(self):
        return self.response 

class TestDiscordTaskSuite(TransactionTestCase):
    def setUp(self):
        DiscordClient.objects.create(
            callback_url="https://localhost:8000",
            server_id="1",
            client_id="1",
            client_secret="null",
            bot_token="null",
            invite_link="https://localhost",
        )
        self.group = Group.objects.create(
            name="Group"
        )

        self.user = User.objects.create(
            username="test"
        )

        self.discord_user = DiscordUser.objects.create(
            username="test",
            nickname="test",
            external_id=1,
        )

        self.no_token_discord_user = DiscordUser.objects.create(
            username="notoken",
            nickname="notoken",
            external_id=2,
        )

        self.discord_token = DiscordToken.objects.create(
            access_token="null",
            refresh_token="null",
            discord_user=self.discord_user,
            user=self.user
        )

        self.discord_group = DiscordGroup.objects.create(
            name="DiscordGroup",
            external_id=1,
            group=self.group
        )

    def tearDown(self):
        User.objects.all().delete()
        Group.objects.all().delete()
        DiscordToken.objects.all().delete()
        DiscordUser.objects.all().delete()
        DiscordGroup.objects.all().delete()

    @patch('django_discord_connector.tasks.remove_discord_user.apply_async')
    @patch('django_discord_connector.tasks.update_discord_user.apply_async')
    @patch('django_discord_connector.tasks.DiscordRequest.remove_role_from_user')
    def test_sync_all_discord_users_accounts(self, mock_discord_call, mock_discord_update_task, mock_discord_remove_task):
        self.assertEqual(DiscordUser.objects.all().count(), 2)
        mock_discord_call.return_value = MockDiscordResponse(status_code=204)
        mock_discord_update_task.return_value = None # assuming update is tested elsewhere
        mock_discord_remove_task.side_effect = remove_discord_user(self.no_token_discord_user.external_id)
        sync_all_discord_users_accounts()
        self.assertEqual(DiscordUser.objects.all().count(), 1)


    @patch('django_discord_connector.tasks.DiscordRequest.remove_role_from_user')
    def test_remove_user(self, mock_discord_call):
        external_id = self.discord_user.external_id
        mock_discord_call.return_value = MockDiscordResponse(status_code=204)
        remove_discord_user(self.discord_user.external_id)
        self.assertFalse(DiscordUser.objects.filter(external_id=external_id).exists())

    @patch('django_discord_connector.tasks.DiscordRequest.remove_role_from_user')
    def test_remove_user_failure_user_still_exists(self, mock_discord_call):
        mock_discord_call.return_value = MockDiscordResponse(status_code=404)
        try:
            remove_discord_user(self.discord_user.external_id)
        except Exception as e:
            pass 
        self.assertTrue(DiscordUser.objects.filter(external_id=self.discord_user.external_id).exists())



from django import forms
from django.forms import ModelForm
from .models import DiscordClient

class DiscordClientForm(ModelForm):
    callback_url = forms.URLField(label="Discord Callback URL")
    server_id = forms.CharField(max_length=255, label="Discord Server ID")
    client_id = forms.CharField(max_length=255, label="Discord Client ID")
    client_secret = forms.CharField(max_length=255, label="Discord Secret ID")
    bot_token = forms.CharField(max_length=255, label="Discord Bot Token")
    invite_link = forms.CharField(max_length=255, label="Discord Invite Link")

    class Meta:
        model = DiscordClient
        fields = ['callback_url', 'server_id', 'client_id', 'client_secret', 'bot_token', 'invite_link']



# Generated by Django 2.2.4 on 2019-09-18 19:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_discord_connector', '0005_auto_20190904_1643'),
    ]

    operations = [
        migrations.AlterField(
            model_name='discorduser',
            name='nickname',
            field=models.CharField(default='Test', max_length=255),
            preserve_default=False,
        ),
    ]



# Generated by Django 2.2.4 on 2019-11-18 14:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_discord_connector', '0006_auto_20190918_1914'),
    ]

    operations = [
        migrations.RenameField(
            model_name='discordclient',
            old_name='secret',
            new_name='client_secret',
        ),
        migrations.AlterField(
            model_name='discordclient',
            name='client_id',
            field=models.TextField(),
        ),
        migrations.AlterField(
            model_name='discordclient',
            name='server_id',
            field=models.TextField(),
        ),
    ]
